{% extends "base.html" %}

{% block content %}
{% for collection_type, collection_list in collections.items() %}
<h3>{{ collection_type }}</h3>
<ul>
  {% for collection in collection_list %}
  <li class="d-flex" style="margin-bottom: 1.5rem;">
    {% with dav_url = "{}{}/".format(BASE_URL, collection.journal.uid) %}
    {{ collection.display_name }}&nbsp;
    <a href="{{ dav_url }}">[link]</a>
    {% endwith %}
  </li>
  {% endfor %}
</ul>
{% else %}
<p>
  No collections found. This should not happen, please contact support.
</p>
{% endfor %}
{% endblock content %}
